import { __assign, __awaiter, __generator, __values } from 'tslib';
import { withTransaction, updateEntities, arrayAdd, arrayRemove, arrayUpdate } from '@datorama/akita';
import { tap, map, finalize } from 'rxjs/operators';
import { race, combineLatest, of } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionService = /** @class */ (function () {
    function CollectionService(db, store, collectionPath) {
        this.db = db;
        this.store = store;
        this.collectionPath = collectionPath;
        if (!this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
    }
    Object.defineProperty(CollectionService.prototype, "idKey", {
        get: /**
         * @return {?}
         */
        function () {
            return this.constructor['idKey'] || this.store.idKey;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path", {
        /** The path to the collection in Firestore */
        get: /**
         * The path to the collection in Firestore
         * @return {?}
         */
        function () {
            return this.constructor['path'] || this.collectionPath;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "collection", {
        get: /**
         * @return {?}
         */
        function () {
            return this.db.collection(this.path);
        },
        enumerable: true,
        configurable: true
    });
    // Helper to retrieve the id and path of a document in the collection
    // Helper to retrieve the id and path of a document in the collection
    /**
     * @private
     * @param {?} __0
     * @return {?}
     */
    CollectionService.prototype.getIdAndPath = 
    // Helper to retrieve the id and path of a document in the collection
    /**
     * @private
     * @param {?} __0
     * @return {?}
     */
    function (_a) {
        var id = _a.id, path = _a.path;
        if (id && path) {
            throw new Error("You should use either the key \"id\" OR \"path\".");
        }
        if (!id && !path) {
            throw new Error("You should provide either an \"id\" OR a \"path\".");
        }
        if (!path) {
            path = this.path + "/" + id;
        }
        if (!id) {
            /** @type {?} */
            var part = path.split('/');
            if (path.length % 2 !== 0) {
                throw new Error("Path " + path + " doesn't look like a Firestore's document path");
            }
            id = part[part.length - 1];
        }
        return { id: id, path: path };
    };
    /** Preformat the document before updating Firestore */
    /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    CollectionService.prototype.preFormat = /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    function (document) {
        return document;
    };
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    CollectionService.prototype.syncCollection = /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    function (pathOrQuery, queryFn) {
        var _this = this;
        if (pathOrQuery === void 0) { pathOrQuery = this.path; }
        /** @type {?} */
        var path;
        if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.path;
        }
        else {
            path = pathOrQuery;
        }
        // Credit to @arielgueta https://dev.to/arielgueta/getting-started-with-akita-and-firebase-3pe2
        /** @type {?} */
        var fromAction = (/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) {
            var e_1, _a, _b;
            _this.store.setLoading(false);
            if (actions.length === 0) {
                return;
            }
            try {
                for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
                    var action = actions_1_1.value;
                    /** @type {?} */
                    var id = (/** @type {?} */ (action.payload.doc.id));
                    /** @type {?} */
                    var entity = action.payload.doc.data();
                    switch (action.type) {
                        case 'added': {
                            _this.store.upsert(id, __assign((_b = {}, _b[_this.idKey] = id, _b), entity));
                            break;
                        }
                        case 'removed': {
                            _this.store.remove(id);
                            break;
                        }
                        case 'modified': {
                            _this.store.update(id, entity);
                        }
                    }
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
        });
        // If there is no entity yet set loading
        if (!this.store._value().ids || this.store._value().ids.length === 0) {
            this.store.setLoading(true);
        }
        // Start Listening
        return (/** @type {?} */ (this.db.collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction(fromAction))));
    };
    /**
     * Stay in sync with one document
     * @param options An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     */
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    CollectionService.prototype.syncDoc = /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    function (options) {
        var _this = this;
        var _a = this.getIdAndPath(options), id = _a.id, path = _a.path;
        // If store doesn't have this value yet, set loading
        if (!this.store._value().ids.includes(id)) {
            this.store.setLoading(true);
        }
        // State listening on change
        return this.db.doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            _this.store.upsert((/** @type {?} */ (id)), __assign({ id: id }, entity));
            _this.store.setLoading(false);
        })));
    };
    /**
     * Stay in sync with the active entity and set it active in the store
     * @param options An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     */
    /**
     * Stay in sync with the active entity and set it active in the store
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    CollectionService.prototype.syncActive = /**
     * Stay in sync with the active entity and set it active in the store
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    function (options) {
        var _this = this;
        return this.syncDoc(options).pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) { return _this.store.setActive((/** @type {?} */ (entity[_this.idKey]))); })));
    };
    /** Return the current value of the path from Firestore */
    /**
     * Return the current value of the path from Firestore
     * @param {?=} id
     * @return {?}
     */
    CollectionService.prototype.getValue = /**
     * Return the current value of the path from Firestore
     * @param {?=} id
     * @return {?}
     */
    function (id) {
        return __awaiter(this, void 0, void 0, function () {
            var snapshot, snapshot;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!id) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.db.doc(this.path + "/" + id).ref.get()];
                    case 1:
                        snapshot = _a.sent();
                        return [2 /*return*/, (/** @type {?} */ (snapshot.data()))];
                    case 2: return [4 /*yield*/, this.collection.ref.get()];
                    case 3:
                        snapshot = _a.sent();
                        return [2 /*return*/, snapshot.docs.map((/**
                             * @param {?} doc
                             * @return {?}
                             */
                            function (doc) { return (/** @type {?} */ (doc.data())); }))];
                }
            });
        });
    };
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     */
    /**
     * Add a document or a list of document to Firestore
     * @param {?} docs A document or a list of document
     * @return {?}
     */
    CollectionService.prototype.add = /**
     * Add a document or a list of document to Firestore
     * @param {?} docs A document or a list of document
     * @return {?}
     */
    function (docs) {
        var _this = this;
        /** @type {?} */
        var addId = (/**
         * @param {?} doc
         * @return {?}
         */
        function (doc) {
            var _a;
            return (__assign({}, doc, (_a = {}, _a[_this.idKey] = doc[_this.idKey] || _this.db.createId(), _a)));
        });
        if (!Array.isArray(docs)) {
            /** @type {?} */
            var doc = addId(docs);
            // We use "set" instead of "add" to use the created id
            return this.db.doc(this.path + "/" + doc[this.idKey]).set(doc);
        }
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        function (tx) {
            return Promise.all(docs.map((/**
             * @param {?} document
             * @return {?}
             */
            function (document) {
                /** @type {?} */
                var doc = addId(_this.preFormat(document));
                var ref = _this.db.doc(_this.path + "/" + doc[_this.idKey]).ref;
                return tx.set(ref, doc);
            })));
        }));
    };
    /** Remove one or several document from Firestore */
    /**
     * Remove one or several document from Firestore
     * @param {?} ids
     * @return {?}
     */
    CollectionService.prototype.remove = /**
     * Remove one or several document from Firestore
     * @param {?} ids
     * @return {?}
     */
    function (ids) {
        var _this = this;
        if (!Array.isArray(ids)) {
            return this.db.doc(this.path + "/" + ids).delete();
        }
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        function (tx) {
            return Promise.all(ids.map((/**
             * @param {?} id
             * @return {?}
             */
            function (id) {
                var ref = _this.db.doc(_this.path + "/" + id).ref;
                return tx.delete(ref);
            })));
        }));
    };
    /**
     * @param {?} idsOrFn
     * @param {?=} newStateOrFn
     * @return {?}
     */
    CollectionService.prototype.update = /**
     * @param {?} idsOrFn
     * @param {?=} newStateOrFn
     * @return {?}
     */
    function (idsOrFn, newStateOrFn) {
        var _this = this;
        /** @type {?} */
        var ids = [];
        // Unique ID : faster than transaction
        if (typeof idsOrFn === 'string') {
            /** @type {?} */
            var id = idsOrFn;
            /** @type {?} */
            var doc = typeof newStateOrFn === 'function'
                ? newStateOrFn(this.store._value().entities[id])
                : newStateOrFn;
            return this.db.doc(this.path + "/" + id).update(doc);
        }
        // Predicate
        if (typeof idsOrFn === 'function') {
            /** @type {?} */
            var state_1 = this.store._value();
            ids = (/** @type {?} */ (state_1.ids.filter((/**
             * @param {?} id
             * @return {?}
             */
            function (id) { return idsOrFn(state_1.entities[id]); }))));
        }
        else {
            ids = idsOrFn;
        }
        if (!Array.isArray(ids) || ids.length === 0) {
            return;
        }
        /** @type {?} */
        var updates = updateEntities({
            state: this.store._value(),
            ids: ids,
            idKey: this.idKey,
            newStateOrFn: newStateOrFn,
            preUpdateEntity: this.store.akitaPreUpdateEntity
        }).entities;
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        function (tx) {
            return Promise.all(Object.keys(updates).map((/**
             * @param {?} key
             * @return {?}
             */
            function (key) {
                /** @type {?} */
                var doc = _this.preFormat(updates[key]);
                var ref = _this.db.doc(_this.path + "/" + doc[_this.idKey]).ref;
                return tx.update(ref, doc);
            })));
        }));
    };
    return CollectionService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options) {
    if (options === void 0) { options = {}; }
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return constructor[key] = options[key]; }));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return (constructor[key] = data[key]); }));
    });
}
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionGuard = /** @class */ (function () {
    function CollectionGuard(service, router) {
        this.service = service;
        this.router = router;
    }
    Object.defineProperty(CollectionGuard.prototype, "awaitSync", {
        // Can be override by the extended class
        /** Should the guard wait for connection to Firestore to complete */
        get: 
        // Can be override by the extended class
        /**
         * Should the guard wait for connection to Firestore to complete
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['awaitSync'] || false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "queryFn", {
        // Can be override by the extended class
        /** A queryFn to execute for sync */
        get: 
        // Can be override by the extended class
        /**
         * A queryFn to execute for sync
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['queryFn'];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "redirect", {
        // Can be override by the extended class
        /** The route to redirect to if you sync failed */
        get: 
        // Can be override by the extended class
        /**
         * The route to redirect to if you sync failed
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['redirect'];
        },
        enumerable: true,
        configurable: true
    });
    // Can be override by the extended class
    /** The method to subscribe to while route is active */
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.sync = 
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _a = (/** @type {?} */ (next.data)).queryFn, queryFn = _a === void 0 ? this.queryFn : _a;
        return this.service.syncCollection(queryFn);
    };
    /**
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.canActivate = /**
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _this = this;
        var _a = (/** @type {?} */ (next.data)), _b = _a.redirect, redirect = _b === void 0 ? this.redirect : _b, _c = _a.awaitSync, awaitSync = _c === void 0 ? this.awaitSync : _c;
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        function (res, rej) {
            if (awaitSync) {
                _this.subscription = _this.sync(next).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    function (result) {
                        switch (typeof result) {
                            case 'string': {
                                _this.subscription.unsubscribe();
                                return res(_this.router.parseUrl(result));
                            }
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    function (err) {
                        _this.subscription.unsubscribe();
                        res(_this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                _this.subscription = _this.sync(next).subscribe();
                res(true);
            }
        }));
    };
    /**
     * @return {?}
     */
    CollectionGuard.prototype.canDeactivate = /**
     * @return {?}
     */
    function () {
        this.subscription.unsubscribe();
        return true;
    };
    return CollectionGuard;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} __0
 * @return {?}
 */
function shoudCancel(_a) {
    var validate = _a.validate, cancel = _a.cancel;
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return false; }))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return true; })))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel(_a) {
    var startWith = _a.startWith, endWith = _a.endWith, shouldValidate = _a.shouldValidate, shouldCancel = _a.shouldCancel;
    return __awaiter(this, void 0, void 0, function () {
        var cancelled;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    startWith();
                    return [4 /*yield*/, race([
                            shouldValidate.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return false; }))),
                            shouldCancel.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return true; })))
                        ]).toPromise()];
                case 1:
                    cancelled = _b.sent();
                    endWith(cancelled);
                    return [2 /*return*/];
            }
        });
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path: path, queryFn: queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path: path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    var _this = this;
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        function (oneQuery) { return syncQuery.bind(_this, oneQuery); })));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    var fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    function (actions, child) {
        var e_1, _a, _b;
        /** @type {?} */
        var idKey = 'id';
        // TODO: Improve how to
        var parentId = child.parentId, key = child.key;
        try {
            for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
                var action = actions_1_1.value;
                /** @type {?} */
                var id = action.payload.doc.id;
                /** @type {?} */
                var data = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        _this['store'].update(parentId, arrayAdd((/** @type {?} */ (key)), __assign((_b = {}, _b[idKey] = id, _b), data)));
                        break;
                    }
                    case 'removed': {
                        _this['store'].update(parentId, arrayRemove((/** @type {?} */ (key)), id, idKey));
                        break;
                    }
                    case 'modified': {
                        _this['store'].update(parentId, arrayUpdate((/** @type {?} */ (key)), id, data));
                    }
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    var syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    function (subQuery, child) {
        var _a;
        // If it's a static value
        if (!isQuery(subQuery)) {
            var parentId = child.parentId, key = child.key;
            /** @type {?} */
            var update = _this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ (_a = {}, _a[key] = subQuery, _a)));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            return combineLatest(subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            function (oneQuery) { return syncSubQuery(oneQuery, child); })));
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            var parentId_1 = child.parentId, key_1 = child.key;
            return _this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            function (children) {
                var _a;
                return _this['store'].update((/** @type {?} */ (parentId_1)), (/** @type {?} */ (_a = {}, _a[key_1] = children, _a)));
            })));
        }
        else {
            return (/** @type {?} */ (_this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            function (actions) { return fromChildAction(actions, child); })))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    var syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    function (subQueries, parent) {
        /** @type {?} */
        var obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return key !== 'path' && key !== 'queryFn'; }))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        function (key) {
            /** @type {?} */
            var queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            var child = { key: key, parentId: parent[_this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    var fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    function (actions, mainQuery, subscriptions) {
        var e_2, _a, _b;
        try {
            for (var actions_2 = __values(actions), actions_2_1 = actions_2.next(); !actions_2_1.done; actions_2_1 = actions_2.next()) {
                var action = actions_2_1.value;
                /** @type {?} */
                var id = action.payload.doc.id;
                /** @type {?} */
                var data = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        /** @type {?} */
                        var entity = __assign((_b = {}, _b[_this.idKey] = id, _b), data);
                        _this['store'].upsert(id, entity);
                        subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                        break;
                    }
                    case 'removed': {
                        _this['store'].remove(id);
                        subscriptions[id].unsubscribe();
                        delete subscriptions[id];
                        break;
                    }
                    case 'modified': {
                        _this['store'].update(id, data);
                    }
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (actions_2_1 && !actions_2_1.done && (_a = actions_2.return)) _a.call(actions_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
    });
    var path = query.path, queryFn = query.queryFn;
    if (isDocPath(path)) {
        var id_1 = this['getIdAndPath']({ path: path }).id;
        /** @type {?} */
        var subscription_1;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            _this['store'].upsert(id_1, (/** @type {?} */ (__assign({ id: id_1 }, entity))));
            if (!subscription_1) { // Subscribe only the first time
                subscription_1 = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        function () { return subscription_1.unsubscribe(); })));
    }
    else {
        /** @type {?} */
        var subscriptions_1 = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return fromActionWithChild(actions, query, subscriptions_1); })), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        function () { return Object.keys(subscriptions_1).forEach((/**
         * @param {?} id
         * @return {?}
         */
        function (id) {
            subscriptions_1[id].unsubscribe();
            delete subscriptions_1[id];
        })); })))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGuard, CollectionGuardConfig, CollectionService, collection, doc, isDocPath, shoudCancel, syncQuery, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
