import { AngularFirestoreCollection, AngularFirestore, QueryFn } from '@angular/fire/firestore';
import { EntityStore, EntityState, UpdateStateCallback, UpdateEntityPredicate, ActiveState, getEntityType } from '@datorama/akita';
import { Observable } from 'rxjs';
import { firestore } from 'firebase';
export declare type CollectionState<E = any> = EntityState<E, string> & ActiveState<string>;
export interface DocOptions {
    path: string;
    id: string;
}
export declare class CollectionService<S extends CollectionState> {
    private db;
    private store;
    private collectionPath?;
    constructor(db: AngularFirestore, store: EntityStore<S>, collectionPath?: string);
    readonly idKey: any;
    /** The path to the collection in Firestore */
    readonly path: any;
    readonly collection: AngularFirestoreCollection<getEntityType<S>>;
    private getIdAndPath;
    /** Preformat the document before updating Firestore */
    protected preFormat<E extends getEntityType<S>>(document: Readonly<E>): E;
    /** Stay in sync with the collection or a fractio of it */
    syncCollection(path?: string | QueryFn): Observable<void>;
    syncCollection(path: string, queryFn?: QueryFn): Observable<void>;
    /**
     * Stay in sync with one document
     * @param options An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     */
    syncDoc(options: Partial<DocOptions>): Observable<getEntityType<S>>;
    /**
     * Stay in sync with the active entity and set it active in the store
     * @param options An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     */
    syncActive(options: Partial<DocOptions>): Observable<getEntityType<S>>;
    /** Return the current value of the path from Firestore */
    getValue(id?: string): Promise<getEntityType<S> | getEntityType<S>[]>;
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     */
    add(docs: getEntityType<S> | getEntityType<S>[]): Promise<void> | Promise<firestore.Transaction[]>;
    /** Remove one or several document from Firestore */
    remove(ids: string | string[]): Promise<void> | Promise<firestore.Transaction[]>;
    /** Update one or several document in Firestore */
    update(id: string, newStateFn: UpdateStateCallback<getEntityType<S>> | Partial<getEntityType<S>>): Promise<void>;
    update(ids: string[] | UpdateEntityPredicate<getEntityType<S>>, newStateFn: UpdateStateCallback<getEntityType<S>> | Partial<getEntityType<S>>): Promise<firestore.Transaction[]>;
}
